#!/usr/bin/env python3
"""
Windows Portable Executable (PE) icon extractor.

.. include:: ../LIB-USAGE.md
"""

import io
import logging
import sys
import struct

import pefile

GRPICONDIRENTRY_FORMAT = ('GRPICONDIRENTRY',
    ('B,Width', 'B,Height','B,ColorCount','B,Reserved',
     'H,Planes','H,BitCount','I,BytesInRes','H,ID'))
GRPICONDIR_FORMAT = ('GRPICONDIR', ('H,Reserved', 'H,Type','H,Count'))

logger = logging.getLogger("icoextract")
logging.basicConfig()

try:
    from .version import __version__
except ImportError:
    __version__ = 'unknown'
    logger.info('icoextract: failed to read program version')

class IconExtractorError(Exception):
    """Superclass for exceptions raised by IconExtractor."""

class NoIconsAvailableError(IconExtractorError):
    """Exception raised when the input program has no icon resources."""

class InvalidIconDefinitionError(IconExtractorError):
    """Exception raised when the input program has an invalid icon resource."""

class IconExtractor():
    def __init__(self, filename=None, data=None):
        """
        Loads an executable from the given `filename` or `data` (raw bytes).
        As with pefile, if both `filename` and `data` are given, `filename` takes precedence.

        If the executable has contains no icons, this will raise `NoIconsAvailableError`.
        """
        # Use fast loading and explicitly load the RESOURCE directory entry. This saves a LOT of time
        # on larger files
        self._pe = pefile.PE(name=filename, data=data, fast_load=True)
        self._pe.parse_data_directories(pefile.DIRECTORY_ENTRY['IMAGE_DIRECTORY_ENTRY_RESOURCE'])

        if not hasattr(self._pe, 'DIRECTORY_ENTRY_RESOURCE'):
            raise NoIconsAvailableError("File has no resources")

        # Reverse the list of entries before making the mapping so that earlier values take precedence
        # When an executable includes multiple icon resources, we should use only the first one.
        resources = {rsrc.id: rsrc for rsrc in reversed(self._pe.DIRECTORY_ENTRY_RESOURCE.entries)}

        self.groupiconres = resources.get(pefile.RESOURCE_TYPE["RT_GROUP_ICON"])
        if not self.groupiconres:
            raise NoIconsAvailableError("File has no group icon resources")
        self.rticonres = resources.get(pefile.RESOURCE_TYPE["RT_ICON"])

    def list_group_icons(self):
        """
        Returns all group icon entries as a list of (name, offset) tuples.
        """
        return [(e.struct.Name, e.struct.OffsetToData)
                for e in self.groupiconres.directory.entries]

    def _get_group_icon_entries(self, num=0):
        """
        Returns the group icon entries for the specified group icon in the executable.
        """
        groupicon = self.groupiconres.directory.entries[num]
        if groupicon.struct.DataIsDirectory:
            # Select the first language from subfolders as needed.
            groupicon = groupicon.directory.entries[0]

        # Read the data pointed to by the group icon directory (GRPICONDIR) struct.
        rva = groupicon.data.struct.OffsetToData
        size = groupicon.data.struct.Size
        data = self._pe.get_data(rva, size)
        file_offset = self._pe.get_offset_from_rva(rva)

        grp_icon_dir = self._pe.__unpack_data__(GRPICONDIR_FORMAT, data, file_offset)
        logger.debug(grp_icon_dir)

        if grp_icon_dir.Reserved:
            raise InvalidIconDefinitionError("Invalid group icon definition (got Reserved=%s instead of 0)" % hex(grp_icon_dir.Reserved))

        # For each group icon entry (GRPICONDIRENTRY) that immediately follows, read its data and save it.
        grp_icons = []
        icon_offset = grp_icon_dir.sizeof()
        for idx in range(grp_icon_dir.Count):
            grp_icon = self._pe.__unpack_data__(GRPICONDIRENTRY_FORMAT, data[icon_offset:], file_offset+icon_offset)
            icon_offset += grp_icon.sizeof()
            grp_icons.append(grp_icon)
            logger.debug("Got logical group icon %s", grp_icon)

        return grp_icons

    def _get_icon_data(self, icon_ids):
        """
        Return a list of raw icon images corresponding to the icon IDs given.
        """
        icons = []
        icon_entry_lists = {icon_entry_list.id: icon_entry_list for icon_entry_list in self.rticonres.directory.entries}
        for icon_id in icon_ids:
            icon_entry_list = icon_entry_lists[icon_id]

            icon_entry = icon_entry_list.directory.entries[0]  # Select first language
            rva = icon_entry.data.struct.OffsetToData
            size = icon_entry.data.struct.Size
            data = self._pe.get_data(rva, size)
            logger.debug(f"Exported icon with ID {icon_entry_list.id}: {icon_entry.struct}")
            icons.append(data)
        return icons

    def _write_ico(self, fd, num=0):
        """
        Writes ICO data to a file descriptor.
        """
        group_icons = self._get_group_icon_entries(num=num)
        icon_images = self._get_icon_data([g.ID for g in group_icons])
        icons = list(zip(group_icons, icon_images))
        assert len(group_icons) == len(icon_images)
        fd.write(b"\x00\x00") # 2 reserved bytes
        fd.write(struct.pack("<H", 1)) # 0x1 (little endian) specifying that this is an .ICO image
        fd.write(struct.pack("<H", len(icons)))  # number of images

        dataoffset = 6 + (len(icons) * 16)
        # First pass: write the icon dir entries
        for datapair in icons:
            group_icon, icon_data = datapair
            # Elements in ICONDIRENTRY and GRPICONDIRENTRY are all the same
            # except the last value, which is an ID in GRPICONDIRENTRY and
            # the offset from the beginning of the file in ICONDIRENTRY.
            fd.write(group_icon.__pack__()[:12])
            fd.write(struct.pack("<I", dataoffset))
            dataoffset += len(icon_data)  # Increase offset for next image

        # Second pass: write the icon data
        for datapair in icons:
            group_icon, icon_data = datapair
            fd.write(icon_data)

    def export_icon(self, filename, num=0):
        """
        Exports ICO data for the requested group icon (`num`) to `filename`.
        """
        with open(filename, 'wb') as f:
            self._write_ico(f, num=num)

    def get_icon(self, num=0):
        """
        Exports ICO data for the requested group icon (`num`) as a `io.BytesIO` instance.
        """
        f = io.BytesIO()
        self._write_ico(f, num=num)
        return f

__all__ = [
    'IconExtractor',
    'IconExtractorError',
    'NoIconsAvailableError',
    'InvalidIconDefinitionError'
]

__pdoc__ = {
    'scripts': False,
    'version': False,
}



#!/usr/bin/env python3

import filecmp
import os.path
import unittest

import icoextract

class UtilsTestCase(unittest.TestCase):
    def _test_extract(self, infile, target):
        # Read/write test files in tests/ folder, regardless of where working directory is
        tests_dir = os.path.dirname(__file__)
        inpath = os.path.join(tests_dir, infile)
        target = os.path.join(tests_dir, target)

        ie = icoextract.IconExtractor(inpath)

        outfile = f"tmp-{infile}.ico"
        outpath = os.path.join(tests_dir, outfile)
        ie.export_icon(outpath)

        self.assertTrue(filecmp.cmp(outpath, target),
                        f"{outpath} and {target} should be equal")
        return ie

    # App has icon + version resource
    def test_testapp64(self):
        ie = self._test_extract("testapp64.exe", "testapp.ico")
        self.assertEqual(len(ie.list_group_icons()), 1)

    # App has only version resource
    def test_testapp64_noicon(self):
        with self.assertRaises(icoextract.NoIconsAvailableError):
            self._test_extract("testapp64-noicon.exe", "testapp-noicon.ico")

    # App has no resource info at all
    def test_testapp64_nores(self):
        with self.assertRaises(icoextract.NoIconsAvailableError):
            self._test_extract("testapp64-nores.exe", "testapp-nores.ico")

    def test_testapp32(self):
        ie = self._test_extract("testapp32.exe", "testapp.ico")
        self.assertEqual(len(ie.list_group_icons()), 1)

    def test_testapp32_noicon(self):
        with self.assertRaises(icoextract.NoIconsAvailableError):
            self._test_extract("testapp32-noicon.exe", "testapp-noicon.ico")

    def test_testapp32_nores(self):
        with self.assertRaises(icoextract.NoIconsAvailableError):
            self._test_extract("testapp32-nores.exe", "testapp-nores.ico")

    def test_fd_as_input(self):
        tests_dir = os.path.dirname(__file__)
        with open(os.path.join(tests_dir, "testapp64.exe"), 'rb') as f:
            ie = icoextract.IconExtractor(data=f.read())
            self.assertEqual(len(ie.list_group_icons()), 1)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
import os.path
import unittest

from icoextract.scripts.thumbnailer import generate_thumbnail
from PIL import Image

TESTS_DIR = os.path.dirname(__file__)
COMPARE_LENGTH = 1024  # for efficiency

class ThumbnailerTestCase(unittest.TestCase):
    def _generate_thumbnail(self, infile, outfile, **kwargs):
        infile_path = os.path.join(TESTS_DIR, infile)
        outfile_path = os.path.join(TESTS_DIR, outfile)
        generate_thumbnail(infile_path, outfile_path, **kwargs)
        return outfile_path

    def _compare_equal(self, im, orig):
        with Image.open(os.path.join(TESTS_DIR, orig)) as im_orig:
            self.assertEqual(list(im.getdata())[:COMPARE_LENGTH], list(im_orig.getdata())[:COMPARE_LENGTH],
                                "Extracted image should match original")

    def test_thumbnailer_normal(self):
        outfile = self._generate_thumbnail("testapp64.exe", "tmp-thumbnail-test-normal.png", size=128)
        with Image.open(outfile) as im:
            self.assertEqual(im.width, 128)
            self.assertEqual(im.height, 128)

    def test_thumbnailer_large(self):
        outfile = self._generate_thumbnail("testapp64.exe", "tmp-thumbnail-test-large.png", size=256)
        with Image.open(outfile) as im:
            self.assertEqual(im.width, 256)
            self.assertEqual(im.height, 256)
            self._compare_equal(im, "testapp.png")

    def test_thumbnailer_with128_large(self):
        outfile = self._generate_thumbnail("testapp64-with128.exe", "tmp-thumbnail-test-with-128-large.png", size=256)
        with Image.open(outfile) as im:
            self.assertEqual(im.width, 256)
            self.assertEqual(im.height, 256)
            self._compare_equal(im, "testapp.png")

    def test_thumbnailer_with128_normal(self):
        outfile = self._generate_thumbnail("testapp64-with128.exe", "tmp-thumbnail-test-with-128-normal.png", size=128)
        with Image.open(outfile) as im:
            self.assertEqual(im.width, 128)
            self.assertEqual(im.height, 128)
            self._compare_equal(im, "tmp-testapp-128.png")

    def test_thumbnailer_smallonly(self):
        outfile = self._generate_thumbnail("testapp32-smallonly.exe", "tmp-thumbnail-test-smallonly.png", size=128)
        with Image.open(outfile) as im:
            self.assertEqual(im.width, 48)
            self.assertEqual(im.height, 48)
            self._compare_equal(im, "tmp-testapp-48.bmp")

    def test_thumbnailer_force_resize(self):
        outfile = self._generate_thumbnail("testapp32-smallonly.exe", "tmp-thumbnail-force-resize.png", size=128,
                                           force_resize=True)
        with Image.open(outfile) as im:
            self.assertEqual(im.width, 128)
            self.assertEqual(im.height, 128)

    def test_192_normal(self):
        """Test that exe files with oddly sized icons (192x192) are wrapped to the expected dimensions"""
        outfile = self._generate_thumbnail("testapp64-with192.exe", "tmp-thumbnail-192-normal.png", size=128)
        with Image.open(outfile) as im:
            self.assertEqual(im.width, 128)
            self.assertEqual(im.height, 128)
            self._compare_equal(im, "tmp-testapp-128.png")

    def test_192_large(self):
        """Test that exe files with oddly sized icons (192x192) are wrapped to the expected dimensions"""
        outfile = self._generate_thumbnail("testapp64-with192.exe", "tmp-thumbnail-192-large.png", size=256)
        with Image.open(outfile) as im:
            self.assertEqual(im.width, 192)
            self.assertEqual(im.height, 192)

    def test_unsupported_output_size_too_large(self):
        """Test an invalid requested icon size (> 256)"""
        outfile = self._generate_thumbnail("testapp64.exe", "tmp-thumbnail-test-unsupported-size-too-large.png",
        size=300)
        with Image.open(outfile) as im:
            self.assertEqual(im.width, 256)
            self.assertEqual(im.height, 256)
            self._compare_equal(im, "testapp.png")

    def test_unsupported_output_size_too_small(self):
        """Test an invalid requested icon size (< 128)"""
        outfile = self._generate_thumbnail("testapp64.exe", "tmp-thumbnail-test-unsupported-size-too-small.png",
        size=64)
        with Image.open(outfile) as im:
            self.assertEqual(im.width, 128)
            self.assertEqual(im.height, 128)
            self._compare_equal(im, "tmp-testapp-128.png")

    def test_unsupported_output_size_between(self):
        """Test an invalid requested icon size (> 128, < 256)"""
        outfile = self._generate_thumbnail("testapp64.exe", "tmp-thumbnail-test-unsupported-size-between.png",
        size=200)
        with Image.open(outfile) as im:
            self.assertEqual(im.width, 128)
            self.assertEqual(im.height, 128)
            self._compare_equal(im, "tmp-testapp-128.png")

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""
Windows PE EXE icon extractor.
"""
import argparse
import logging
import os.path

from icoextract import IconExtractor, logger, __version__

_WRONG_EXTENSIONS_HINT = {'.jpg', '.jpeg', '.png'}
def main():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("-V", "--version", action='version', version=f'icoextract {__version__}')
    parser.add_argument("-n", "--num", type=int, help="index of icon to extract", default=0)
    parser.add_argument("-v", "--verbose", action="store_true", help="enables debug logging")
    parser.add_argument("input", help="input filename (.exe/.dll/.mun)")
    parser.add_argument("output", help="output filename (.ico)")
    args = parser.parse_args()

    if args.verbose:
        logger.setLevel(logging.DEBUG)

    extractor = IconExtractor(args.input)
    extractor.export_icon(args.output, num=args.num)
    file_ext = os.path.splitext(args.output)[1].lower()
    if file_ext in _WRONG_EXTENSIONS_HINT:
        logger.warning('This tool outputs .ico files, not %s. The resulting file will have the wrong file extension.',
                       file_ext)



#!/usr/bin/env python3
"""
Linux (freedesktop.org) thumbnailer for Windows PE files (.exe/.dll)
"""
import argparse
import logging
import sys

from PIL import Image

from icoextract import IconExtractor, logger, __version__

def generate_thumbnail(inputfile, outfile, size=256, force_resize=False):
    """
    Generates a thumbnail for an .exe file.

    inputfile: the input file path (%i)
    outfile: output filename (%o)
    size: determines the thumbnail output size (%s)
    """
    try:
        extractor = IconExtractor(inputfile)
    except RuntimeError:
        logger.debug("Failed to extract icon for %s:", inputfile, exc_info=True)
        sys.exit(1)
    data = extractor.get_icon()

    im = Image.open(data)  # Open up the .ico from memory
    if force_resize:
        logger.debug("Force resizing icon to %dx%d", size, size)
        im = im.resize((size, size))
    else:
        if size > 256:
            logger.warning('Icon sizes over 256x256 are not supported')
            size = 256
        elif size not in (128, 256):
            logger.warning('Unsupported size %d, falling back to 128x128', size)
            size = 128

        # Note: 256x256 is the largest size supported by the .ico format
        if size == 256:
            # A large size thumbnail was requested. No downwards resizing is needed, so export any icon as is
            logger.debug("Writing large size thumbnail for %s to %s", inputfile, outfile)
            im.save(outfile, "PNG")
            return

        # If large size thumbnail wasn't requested but one is available, pick an 128x128 icon if available;
        # otherwise scale down from 256x256 to 128x128. 128x128 is the largest resolution allowed for
        # "normal" size thumbnails.
        if (128, 128) in im.info['sizes']:
            logger.debug("Using native 128x128 icon")
            im.size = (128, 128)
        elif im.size > (128, 128):
            logger.debug("Downsizing icon to 128x128")
            im = im.resize((128, 128))
        logger.debug("Writing normal size thumbnail for %s to %s", inputfile, outfile)

    im.save(outfile, "PNG")

def main():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("-V", "--version", action='version', version=f'exe-thumbnailer, part of icoextract {__version__}')
    parser.add_argument("-s", "--size", type=int, help="size of desired thumbnail", default=256)
    parser.add_argument("-v", "--verbose", action="store_true", help="enables debug logging")
    parser.add_argument("-f", "--force-resize", action="store_true", help="force resize thumbnail to the specified size")
    parser.add_argument("inputfile", help="input file name (.exe/.dll/.mun)")
    parser.add_argument("outfile", help="output file name (.png)")
    args = parser.parse_args()

    if args.verbose:
        logger.setLevel(logging.DEBUG)

    generate_thumbnail(args.inputfile, args.outfile, size=args.size, force_resize=args.force_resize)



#!/usr/bin/env python3

from setuptools import setup, find_packages

with open('icoextract/version.py') as f:
    exec(f.read())

setup(
    name="icoextract",
    description="Windows PE EXE icon extractor",
    version=__version__,
    url="https://github.com/jlu5/icoextract",

    author="James Lu",
    author_email="james@overdrivenetworks.com",

    license="MIT/Expat",
    classifiers=[
        # https://pypi.org/classifiers/
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],

    packages=find_packages(exclude=['tests']),
    install_requires=['pefile'],
    extras_require={
        "thumbnailer": ["Pillow"]
    },

    # Executable scripts
    entry_points={
        'console_scripts': [
            'icoextract = icoextract.scripts.extract:main',
            'icolist = icoextract.scripts.icolist:main',
            'exe-thumbnailer = icoextract.scripts.thumbnailer:main [thumbnailer]',
        ],
    },
)



#!/usr/bin/env python3
"""
Lists group icons present in a program.
"""
import argparse
import logging

from icoextract import IconExtractor, logger, __version__

def main():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("-V", "--version", action='version', version=f'icoextract {__version__}')
    parser.add_argument("-v", "--verbose", action="store_true", help="enables debug logging")
    parser.add_argument("input", help="input filename")
    args = parser.parse_args()

    if args.verbose:
        logger.setLevel(logging.DEBUG)

    extractor = IconExtractor(args.input)
    for idx, entry in enumerate(extractor.list_group_icons()):
        eid, offset = entry
        print(f"Index: {idx}    "
              f"ID: {eid}({hex(eid)})    "
              f"Offset: {hex(offset)}")
